import bpy

# -----------------------------------------------------------------------------
#
bl_info = {
    "name": "Turok 2 EX Tools",
    "description": "Import and export models and animations",
    "author": "BehemothProgrammer",
    "version": (1, 5, 0),
    "blender": (2, 82, 7),
    "location": "File > Import-Export",
    "wiki_url": "https://behemothprogrammer.github.io/Turok2Blender/doc.html",
    "support": "COMMUNITY",
    "category": "Import-Export"}
  
# -----------------------------------------------------------------------------
#
if "bpy" in locals():
    import importlib
    if "paths" in locals():
        importlib.reload(paths)
    if "utils" in locals():
        importlib.reload(utils)
    if "file_utils" in locals():
        importlib.reload(file_utils)
    if "panel_preferences" in locals():
        importlib.reload(panel_preferences)
    if "panel_object" in locals():
        importlib.reload(panel_object)
    if "handlers" in locals():
        importlib.reload(handlers)
    if "staticmesh" in locals():
        importlib.reload(staticmesh)
    if "kmaterial" in locals():
        importlib.reload(kmaterial)
    if "kdefs" in locals():
        importlib.reload(kdefs)
    if "panel_material" in locals():
        importlib.reload(panel_material)
    if "panel_shading" in locals():
        importlib.reload(panel_shading)
    if "panel_tool" in locals():
        importlib.reload(panel_tool)
    if "aud_utils" in locals():
        importlib.reload(aud_utils)
    if "skinnedmesh" in locals():
        importlib.reload(skinnedmesh)
    if "panel_bone" in locals():
        importlib.reload(panel_bone)
    if "animation" in locals():
        importlib.reload(animation)
    if "panel_action" in locals():
        importlib.reload(panel_action)
#end if

from datetime import (
    datetime,
    timedelta,
)

from . import paths
from . import utils
from . import file_utils
from . import panel_preferences
from . import panel_object
from . import handlers
from . import staticmesh
from . import kmaterial
from . import kdefs
from . import panel_material
from . import panel_shading
from . import panel_tool
from . import aud_utils
from . import skinnedmesh
from . import panel_bone
from . import animation
from . import panel_action

# -----------------------------------------------------------------------------
# Called first thing when register is called
def OnStart():
    utils.TUROK2_BL_INFO = bl_info
    kdefs.OnStart()
#end def

# -----------------------------------------------------------------------------
#
def register():
    OnStart()
    
    panel_object.register()
    panel_preferences.register()
    handlers.register()
    staticmesh.register()
    skinnedmesh.register()
    animation.register()
    panel_material.register()
    panel_shading.register()
    panel_tool.register()
    panel_bone.register()
    panel_action.register()

    # print("")
    # print("------------------Start Program---------------")
    # print("")
    
    # import sys
    # import os
    # print(paths.addon_dir())
    # print(__file__) #path to this file
    # print(bpy.app.version) #this addon version
    # print(__package__) #package name "io_scene_kex_turok"
    # print(sys.version_info) #python version
    
    # print("")
    # print("------------------End Program---------------")
    # print("")
#end def
    
# -----------------------------------------------------------------------------
#
def unregister():
    panel_object.unregister()
    panel_preferences.unregister()
    handlers.unregister()
    staticmesh.unregister()
    skinnedmesh.unregister()
    animation.unregister()
    panel_material.unregister()
    panel_shading.unregister()
    panel_tool.unregister()
    panel_bone.unregister()
    panel_action.unregister()
#end def

# -----------------------------------------------------------------------------
#
if __name__ == "__main__":
    register()
#end if
