import bpy
import os

from mathutils import (
    Vector,
    Quaternion,
    Matrix,
)

from . import (
    utils,
    aud_utils,
    kdefs,
    panel_object,
    panel_tool,
)

HIDE_ARG_NAMES = {"Arg 1", "Arg 2", "Arg 3", "Arg 4", "Arg 5", "Arg 6", "Arg 7", "Arg 8", "Unused"}
lastKFAClickTime = 0.0
# -----------------------------------------------------------------------------
#
class Turok2ActionKeyFramesPropGroup(bpy.types.PropertyGroup):
    def on_event_changed(self, context):
        self.name = kdefs.get_anim_kfa_fullname(self.event, self.get_all_args())
    #end def
        
    def get_all_args(self):
        return [self.arg1, self.arg2, self.arg3, self.arg4, self.arg5, self.arg6, self.arg7, self.arg8]
    #end def
    
    def get_args(self, argIndex):
        if argIndex == 0:
            return self.arg1
        elif argIndex == 1:
            return self.arg2
        elif argIndex == 2:
            return self.arg3
        elif argIndex == 3:
            return self.arg4
        elif argIndex == 4:
            return self.arg5
        elif argIndex == 5:
            return self.arg6
        elif argIndex == 6:
            return self.arg7
        else:
            return self.arg8
        #end if
    #end def
            
    def set_args(self, argIndex, value):
        if argIndex == 0:
            self.arg1 = value
        elif argIndex == 1:
            self.arg2 = value
        elif argIndex == 2:
            self.arg3 = value
        elif argIndex == 3:
            self.arg4 = value
        elif argIndex == 4:
            self.arg5 = value
        elif argIndex == 5:
            self.arg6 = value
        elif argIndex == 6:
            self.arg7 = value
        else:
            self.arg8 = value
        #end if
    #end def
            
    def get_arg_names(self):
        argNames = [""] * 8
        argValues = self.get_all_args()
        for i in range(8):
            argName = kdefs.get_anim_kfa_argname(self.event, i, argValues)
            if not argName:
                argName = "Arg %i" % (i+1)
            #end if
            argNames[i] = argName
        #end for
        
        return argNames
    #end def
        
    bone: bpy.props.IntProperty(name="Bone", min=0, max=65535, default=65535, description="This is not used")
    offset: bpy.props.FloatVectorProperty(name="Offset", description="Offset relative to Models Origin (Event 11: Fire Weapon Particle does NOT use this.)", default=Vector((0.0, 0.0, 0.0)), subtype='XYZ', size = 3, step=100, precision=6)
    event: bpy.props.IntProperty(name="Event", min=0, max=65535, update=on_event_changed)
    frame: bpy.props.IntProperty(name="Frame", min=0, max=65535, default=0)
    arg1: bpy.props.FloatProperty(name="Arg1", step=100, precision=6)
    arg2: bpy.props.FloatProperty(name="Arg2", step=100, precision=6)
    arg3: bpy.props.FloatProperty(name="Arg3", step=100, precision=6)
    arg4: bpy.props.FloatProperty(name="Arg4", step=100, precision=6)
    arg5: bpy.props.FloatProperty(name="Arg5", step=100, precision=6)
    arg6: bpy.props.FloatProperty(name="Arg6", step=100, precision=6)
    arg7: bpy.props.FloatProperty(name="Arg7", step=100, precision=6)
    arg8: bpy.props.FloatProperty(name="Arg8", step=100, precision=6)
#end def

# -----------------------------------------------------------------------------
#
class TUROK2_UL_key_frames_props(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        row = layout.row(align=True)
        row = row.split(factor=0.85)
        row.label(text="%s" % (kdefs.get_anim_kfa_fullname(item.event, item.get_all_args())))
        row.label(text="(%d)" % (item.frame))
    #end def
#end class

# -----------------------------------------------------------------------------
#
class Turok2AnimationActionSettings(bpy.types.PropertyGroup):

    def on_get_animID(self):
        return self.get("animID", -1)
    #end if
    def on_set_animID(self, value):
        self["animID"] = value
        
        if bpy.context is None or not hasattr(bpy.context, "object"):
            return
        #end if
        obj = bpy.context.object
        if obj is None:
            return
        #end if
        armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.t2.objType != "SKINNEDMESH" or armObj.animation_data is None or armObj.animation_data.action is None:
            return
        #end if
        
        animName = kdefs.defs["anims"].get(value)
        if animName:
            armObj.animation_data.action.name = animName
        #end if
    #end if
    
    def on_keyframesindex_changed(self, context):
        global lastKFAClickTime
        
        if self.keyFramesIndex >= 0 and self.keyFramesIndex < len(self.keyFrames):
            if utils.time_reached(lastKFAClickTime):
                lastKFAClickTime = utils.time_from(utils.DOUBLE_CLICK_TIME)
            else:
                #if that key frame has an arg1 of soundID or voiceID then play the sound/voice if clicked within x amount of time
                lastKFAClickTime = utils.time_from()
                keyFrame = self.keyFrames[self.keyFramesIndex]
                argNames = keyFrame.get_arg_names()
                if argNames[0] == "Sound ID":
                    aud_utils.stop_sound()
                    sound = kdefs.get_sound(keyFrame.arg1, -1)
                    aud_utils.play_ksnd(sound)
                elif argNames[0] == "Voice ID":
                    aud_utils.stop_sound()
                    aud_utils.play_voice(keyFrame.arg1)
                #end if
            #end if
        #end if
    #end def
    
    keyFrames: bpy.props.CollectionProperty(name="Key Frame Actions", type=Turok2ActionKeyFramesPropGroup)
    keyFramesIndex: bpy.props.IntProperty(name="", default = -1, update=on_keyframesindex_changed)
    frameScale: bpy.props.FloatProperty(name="Frame Scale", description="Unused", step=100, precision=6, default=1.0)
    blendLength: bpy.props.IntProperty(name="Blend Length", description="Length of frames to blend", default=0, min=0, max=65535)
    blendStart: bpy.props.IntProperty(name="Blend Start", default=0, min=0, max=65535)
    blendEnd: bpy.props.IntProperty(name="Blend End", default=0, min=0, max=65535)
    speed: bpy.props.FloatProperty(name="FPS", description="Frames per second for this animation", step=100, precision=6, default=15.0, min=1.0)
    loopFrame: bpy.props.IntProperty(name="Loop Frame", default=0, min=0, max=65535, description="start frame after looping")
    animID: bpy.props.IntProperty(name="AnimID", min=-2147483648, max=2147483647, default=-1, get=on_get_animID, set=on_set_animID)
    
    def copy_keyframe(self, index):
        if index < 0 or index >= len(self.keyFrames):
            return
        #end if
        keyFrame = self.keyFrames[index]
        utils.copiedKeyFrame = [keyFrame.frame, keyFrame.event, keyFrame.bone, keyFrame.offset, keyFrame.arg1, keyFrame.arg2, keyFrame.arg3, keyFrame.arg4, keyFrame.arg5, keyFrame.arg6, keyFrame.arg7, keyFrame.arg8]
    #end def
        
    def paste_keyframe(self, index):
        if utils.copiedKeyFrame is None or index < 0 or index >= len(self.keyFrames):
            return
        #end if
        self.keyFrames[index].frame = utils.copiedKeyFrame[0]
        self.keyFrames[index].bone = utils.copiedKeyFrame[2]
        self.keyFrames[index].offset = utils.copiedKeyFrame[3]
        self.keyFrames[index].arg1 = utils.copiedKeyFrame[4]
        self.keyFrames[index].arg2 = utils.copiedKeyFrame[5]
        self.keyFrames[index].arg3 = utils.copiedKeyFrame[6]
        self.keyFrames[index].arg4 = utils.copiedKeyFrame[7]
        self.keyFrames[index].arg5 = utils.copiedKeyFrame[8]
        self.keyFrames[index].arg6 = utils.copiedKeyFrame[9]
        self.keyFrames[index].arg7 = utils.copiedKeyFrame[10]
        self.keyFrames[index].arg8 = utils.copiedKeyFrame[11]
        self.keyFrames[index].event = utils.copiedKeyFrame[1]
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_loop_frame_last(bpy.types.Operator):
    bl_idname = "turok2_action.loopframe_last"
    bl_label = "Loop Frame to last"
    bl_description = "Sets the loop frame to the last frame"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        return armObj.type == "ARMATURE" and armObj.animation_data is not None and armObj.animation_data.action is not None
    #end def
        
    def invoke(self, context, event):
        obj = utils.get_root_object(context.object, "ARMATURE")
        action = obj.animation_data.action
        frameStart, frameEnd = action.frame_range
        frameStart = int(frameStart)
        frameEnd = int(frameEnd)
        frameCount = (frameEnd - frameStart) + 1
        action.t2.loopFrame = frameCount - 1
                
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_key_frame_action(bpy.types.Operator):
    bl_idname = "turok2_action.keyframe_action"
    bl_label = "Perform Key Frame UI Action"
    bl_description = "Move items up and down, add and remove"
    bl_options = {'INTERNAL', 'UNDO'}
    
    action: bpy.props.EnumProperty(items=(  ('UP', "Up", ""),
                                            ('DOWN', "Down", ""),
                                            ('REMOVE', "Remove", ""),
                                            ('ADD', "Add", "")
                                         ))
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        return armObj.type == "ARMATURE" and armObj.animation_data is not None and armObj.animation_data.action is not None
    #end def
        
    def invoke(self, context, event):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        idx = tAction.keyFramesIndex
        if self.action == 'DOWN' and idx < len(tAction.keyFrames) - 1:
            tAction.keyFrames.move(idx, idx + 1)
            tAction.keyFramesIndex += 1
        elif self.action == 'UP' and idx >= 1:
            tAction.keyFrames.move(idx, idx - 1)
            tAction.keyFramesIndex -= 1
        elif self.action == 'REMOVE' and idx >= 0 and idx < len(tAction.keyFrames):
            tAction.keyFrames.remove(idx)
            if idx >= len(tAction.keyFrames):
                tAction.keyFramesIndex -= 1
            #end if
        elif self.action == 'ADD':
            item = tAction.keyFrames.add()
            item.name = kdefs.get_anim_kfa_fullname(item.event, item.get_all_args())
            tAction.keyFramesIndex = len(tAction.keyFrames) - 1
        #end if
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_key_frame_copy(bpy.types.Operator):
    bl_idname = "turok2_action.keyframe_copy"
    bl_label = "Key Frame Action Copy"
    bl_description = "Copy Key Frame Action"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def
    
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        tAction.copy_keyframe(tAction.keyFramesIndex)
            
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_key_frame_paste(bpy.types.Operator):
    bl_idname = "turok2_action.keyframe_paste"
    bl_label = "Key Frame Action Paste"
    bl_description = "Paste Key Frame Action"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None or utils.copiedKeyFrame is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def
    
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        tAction.paste_keyframe(tAction.keyFramesIndex)
            
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_key_frame_set_anim_frame(bpy.types.Operator):
    bl_idname = "turok2_action.keyframe_set_anim_frame"
    bl_label = "Set scene animation frame"
    bl_description = "Sets the scene animation frame to the Key Frame Action Frame"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def
    
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        kfa = tAction.keyFrames[tAction.keyFramesIndex]
        bpy.context.scene.frame_set(kfa.frame)
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_key_frame_toggle_point(bpy.types.Operator):
    bl_idname = "turok2_action.keyframe_toggle_point"
    bl_label = "Toggle Key Frame Action Point"
    bl_description = "Show or Hide the Key Frame Action Point"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        kfa = tAction.keyFrames[tAction.keyFramesIndex]
        turok2PointObj = utils.get_object(utils.TUROK2_ACTIONPOINT_NAME)
        if turok2PointObj is None:
            pointObj = utils.create_kfa_point(context, kfa, obj)
            pointObj.select_set(True)
        else:
            bpy.data.objects.remove(turok2PointObj)
        #end if
            
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_key_frame_point_update(bpy.types.Operator):
    bl_idname = "turok2_action.keyframe_point_update"
    bl_label = "Update Key Frame Action Point"
    bl_description = "Updates the Key Frame Action Point object's location and scale using the Args"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        turok2PointObj = utils.get_object(utils.TUROK2_ACTIONPOINT_NAME)
        if turok2PointObj is None:
            return False
        #end if
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        kfa = tAction.keyFrames[tAction.keyFramesIndex]
        pointObj = utils.create_kfa_point(context, kfa, obj)
        pointObj.select_set(True)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_key_frame_point_to_args(bpy.types.Operator):
    bl_idname = "turok2_action.keyframe_point_to_args"
    bl_label = "Set Key Frame Action Args to Point"
    bl_description = "Sets the Key Frame Action Args to the point objects location and scale"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        turok2PointObj = utils.get_object(utils.TUROK2_ACTIONPOINT_NAME)
        if turok2PointObj is None:
            return False
        #end if
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def
    
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        kfa = tAction.keyFrames[tAction.keyFramesIndex]
        argNames = kfa.get_arg_names()
        radiusArgIndex = utils.index_of(argNames, "Radius")
        pointObj = utils.get_object(utils.TUROK2_ACTIONPOINT_NAME)
        kfa.offset = Vector((-pointObj.location[0], -pointObj.location[1], pointObj.location[2]))
        if radiusArgIndex != -1:
            kfa.set_args(radiusArgIndex, (pointObj.scale[0] / utils.TUROK2_KFA_POINT_RADIUS_SCALE) * 256.0)
        #end if
            
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_MT_key_frame_special_menu(bpy.types.Menu):
    bl_idname = "TUROK2_MT_key_frame"
    bl_label = "keyframe menu"
    
    def draw(self, context):
        layout = self.layout
        layout.operator(TUROK2_OT_key_frame_copy.bl_idname, icon='COPYDOWN', text="Copy")
        layout.operator(TUROK2_OT_key_frame_paste.bl_idname, icon='PASTEDOWN', text="Paste")
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_action_bone_list(bpy.types.Operator):
    bl_idname = "turok2_action.actionbone_list"
    bl_label = "KFA Bone"
    bl_description = "Select a Armature Bone from a list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        index = 0
        armObj = utils.get_root_object(context.object, "ARMATURE")
        armature = armObj.data
        for bone in armature.bones:
            items.append((bone.name, bone.name, "", index))
            index += 1
        #end for
        return items
    #end def
            
    enumprop: bpy.props.EnumProperty(items=items_cb)

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def
        
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        armature = obj.data
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        boneName = self.enumprop
        if boneName in armature.bones:
            keyFrame.bone = armature.bones[boneName].t2.boneIndex
        #end if
        
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_event_list(bpy.types.Operator):
    """Select a Turok 2 Animation Event from a list"""
    bl_idname = "turok2_action.animevent_list"
    bl_label = "Set Turok Anim Event with list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["anim_events"]:
            actionEvent = kdefs.defs["anim_events"][id]
            items.append((str(id), actionEvent["name"], "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def
        
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        keyFrame.event = utils.get_enum_type(self.items_cb(context), self.enumprop)
        
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_action_arg_sound_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a sound list"""
    bl_idname = "turok2_action.animaction_arg_sound_list"
    bl_label = "Set Turok 2 Anim Action Arg with sound list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["sounds"]:
            for sound in kdefs.defs["sounds"][id]:
                name = os.path.splitext(os.path.split(sound["path"])[1])[0]
                items.append(("%s %s" % (id, sound["index"]), name, "", len(items)))
            #end for
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        enumData = self.enumprop.split()
        soundID = int(enumData[0])
        keyFrame.set_args(self.argIndex, soundID)
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_action_arg_voice_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a voice list"""
    bl_idname = "turok2_action.animaction_arg_voice_list"
    bl_label = "Set Turok 2 Anim Action Arg with voice list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["voices"]:
            voice = kdefs.defs["voices"][id]
            name = os.path.split(voice["name"])[1]
            items.append((str(id), name, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        keyFrame.set_args(self.argIndex, utils.get_enum_type(self.items_cb(context), self.enumprop))
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_action_arg_particle_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a particle list"""
    bl_idname = "turok2_action.animaction_arg_particle_list"
    bl_label = "Set Turok 2 Anim Action Arg with particle list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["particles"]:
            particle = kdefs.defs["particles"][id]
            name = os.path.splitext(os.path.split(particle["path"])[1])[0]
            items.append((str(id), name, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        keyFrame.set_args(self.argIndex, utils.get_enum_type(self.items_cb(context), self.enumprop))
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_action_arg_actor_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a actor list"""
    bl_idname = "turok2_action.animaction_arg_actor_list"
    bl_label = "Set Turok 2 Anim Action Arg with actor list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["actors"]:
            actor = kdefs.defs["actors"][id]
            name = actor["name"]
            items.append((str(id), name, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        keyFrame.set_args(self.argIndex, utils.get_enum_type(self.items_cb(context), self.enumprop))
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_action_arg_hotpoint_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a hotpoint list"""
    bl_idname = "turok2_action.animaction_arg_hotpoint_list"
    bl_label = "Set Turok 2 Anim Action Arg with hotpoint list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["hotpoints"]:
            name = kdefs.defs["hotpoints"][id]
            items.append((str(id), name, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        keyFrame.set_args(self.argIndex, utils.get_enum_type(self.items_cb(context), self.enumprop))
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_action_arg_impact_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a impact list"""
    bl_idname = "turok2_action.animaction_arg_impact_list"
    bl_label = "Set Turok 2 Anim Action Arg with impact list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["impacts"]:
            name = kdefs.defs["impacts"][id]
            items.append((str(id), name, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        keyFrame.set_args(self.argIndex, utils.get_enum_type(self.items_cb(context), self.enumprop))
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_list(bpy.types.Operator):
    """Set Turok 2 Animation ID from a list"""
    bl_idname = "turok2_action.anim_list"
    bl_label = "Set Turok 2 Anim ID with a list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["anims"]:
            animName = kdefs.defs["anims"][id]
            items.append((str(id), animName, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        return True
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        tAction.animID = utils.get_enum_type(self.items_cb(context), self.enumprop)
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_anim_action_arg_userevent_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a custom user event list"""
    bl_idname = "turok2_action.animaction_arg_userevent_list"
    bl_label = "Set Turok 2 Anim Action Arg with custom user event list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        listInfo = kdefs.defs["user_events"]["lists"].get(self.listName)
        items = []
        for id in listInfo:
            name = listInfo[id]
            items.append((str(id), name, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()
    listName: bpy.props.StringProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        keyFrame.set_args(self.argIndex, utils.get_enum_type(self.items_cb(context), self.enumprop))
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
# custom list of events for arg1
class TUROK2_OT_anim_action_arg_userevent1_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a custom user event list"""
    bl_idname = "turok2_action.animaction_arg_userevent1_list"
    bl_label = "Set Turok 2 Anim Action Arg1 with custom user event list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["user_events"]["events"]:
            eventInfo = kdefs.defs["user_events"]["events"][id]
            items.append((str(id), eventInfo["name"], "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    argIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def

    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        keyFrame = tAction.keyFrames[tAction.keyFramesIndex]
        keyFrame.set_args(self.argIndex, utils.get_enum_type(self.items_cb(context), self.enumprop))
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_key_frame_set_bone_none(bpy.types.Operator):
    bl_idname = "turok2_action.keyframe_set_bone_none"
    bl_label = "Set None"
    bl_description = "Sets the Key Frame Action Bone to none"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return False
        #end if
        tAction = armObj.animation_data.action.t2
        return tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames)
    #end def
    
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        tAction = obj.animation_data.action.t2
        kfa = tAction.keyFrames[tAction.keyFramesIndex]
        kfa.bone = 0xFFFF
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_PT_action(bpy.types.Panel):
    bl_label = "Turok 2 Action"
    bl_idname = "TUROK2_PT_action"
    bl_space_type = 'DOPESHEET_EDITOR'
    bl_region_type = 'UI'
    bl_context = "action"

    def draw(self, context):
        obj = context.object
        if obj is None or obj.animation_data is None or obj.animation_data.action is None:
            return
        #end if
            
        action = obj.animation_data.action
        turok2_draw_action_panel(self, context, action)
    #end def
#end class

# -----------------------------------------------------------------------------
#
def turok2_draw_action_panel(self, context, action):
    obj = utils.get_root_object(context.object, "ARMATURE")
    if obj.type != "ARMATURE":
        return
    #end if
    
    layout = self.layout
    tAction = action.t2
    prefs = utils.prefs()

    row = layout.row()
    row.prop(obj.t2, "actionTabIndex", expand=True)
    box = layout.box()
    row = box.row()
    row.label(text="Anim %d: %s" % (tAction.animID, action.name), icon="ANIM", translate=False)
    if obj.t2.actionTabIndex == "DATA":
        row = box.row(align=True)
        row.prop(tAction, "animID")
        row.label(text=kdefs.defs["anims"].get(tAction.animID, "Anim %i" % tAction.animID))
        row.operator(TUROK2_OT_anim_list.bl_idname, icon='COLLAPSEMENU', text="")
        row = box.row()
        row.prop(tAction, "speed")
        row = box.row()
        row.prop(tAction, "frameScale")
        row = box.row()
        row.prop(tAction, "blendLength")
        row = box.row()
        row.prop(tAction, "blendStart")
        row = box.row()
        row.prop(tAction, "blendEnd")
        row = box.row(align=True)
        row.prop(tAction, "loopFrame")
        row.operator(TUROK2_OT_loop_frame_last.bl_idname, icon='FRAME_NEXT', text="")
    else: #KFA
        row = box.row()
        keyFramesText = 'Action Key Frames'
        if len(tAction.keyFrames) > 0:
            keyFramesText = 'Action Key Frames (%d)' % (len(tAction.keyFrames))
        row.label(text=keyFramesText)
        row = box.row()
        row.template_list('TUROK2_UL_key_frames_props', "Turok2ActionKeyFramesPropGroup", tAction, "keyFrames", tAction, "keyFramesIndex", type='DEFAULT')
        col = row.column(align=True)
        col.operator(TUROK2_OT_key_frame_action.bl_idname, icon='ADD', text="").action = 'ADD'
        col.operator(TUROK2_OT_key_frame_action.bl_idname, icon='REMOVE', text="").action = 'REMOVE'
        col.separator()
        col.menu(TUROK2_MT_key_frame_special_menu.bl_idname, icon='DOWNARROW_HLT', text="")
        col.separator()
        col.operator(TUROK2_OT_key_frame_action.bl_idname, icon='TRIA_UP', text="").action = 'UP'
        col.operator(TUROK2_OT_key_frame_action.bl_idname, icon='TRIA_DOWN', text="").action = 'DOWN'
        
        if tAction.keyFramesIndex >= 0 and tAction.keyFramesIndex < len(tAction.keyFrames):
            item = tAction.keyFrames[tAction.keyFramesIndex]
            argNames = item.get_arg_names()
            
            #Data (Frame and Bone Index)
            titleRow = box.box().row()
            titleRow.alignment = "CENTER"
            titleRow.label(text="Data", icon="PMARKER_ACT")
            
            row = box.row(align=True)
            row.prop(item, "frame")
            row.operator(TUROK2_OT_key_frame_set_anim_frame.bl_idname, icon='FORWARD', text="")
            
            #Bone Index
            boneName = ""
            if item.bone != 0xFFFF:
                for bone in obj.data.bones:
                    if bone.t2.boneIndex == item.bone:
                        boneName = bone.name
                        break
                    #end if
                #end for
            #end if
            row = box.row(align=True)
            row.alert = item.bone != 0xFFFF and not boneName
            row.prop(item, "bone")
            if item.bone == 0xFFFF:
                row.label(text="None")
                row.operator(TUROK2_OT_action_bone_list.bl_idname, icon='COLLAPSEMENU', text="")
            else:
                row.label(text=boneName)
                row.operator(TUROK2_OT_action_bone_list.bl_idname, icon='COLLAPSEMENU', text="")
                row.operator(TUROK2_OT_key_frame_set_bone_none.bl_idname, icon='X', text="")
            #end if

            #Event
            box.separator()
            titleBox = box.box()
            titleRow = titleBox.row(align=True)
            titleRow.alignment = "CENTER"
            titleRow.label(text="Event", icon="EVENT_E")
            
            row = box.row(align=True)
            animEventDef = kdefs.defs["anim_events"].get(item.event)
            row.prop(item, "event", text=(animEventDef["name"] if animEventDef else "Event ID"))
            row.operator(TUROK2_OT_anim_event_list.bl_idname, icon='COLLAPSEMENU', text="")
            box.separator()

            #Args
            itemArgs = item.get_all_args()
            for i in range(8):
                argDesc = argNames[i]#utils.get_anim_kfa_argname(item.event, i, itemArgs)
                if prefs.hideKFAArgs and argDesc in HIDE_ARG_NAMES:
                    continue
                #end if
                col = box.column(align=True)
                row = col.row(align=True)
                row.prop(item, "arg%i" % (i+1), text=argNames[i])
                argText = kdefs.get_anim_kfa_arg_lookup(item.event, i, itemArgs)
                if argDesc == "Sound ID":
                    row.operator(TUROK2_OT_anim_action_arg_sound_list.bl_idname, icon='COLLAPSEMENU', text="").argIndex = i
                    row = col.row(align=True)
                    row.alignment = "RIGHT"
                    row.label(text=argText)
                    opPlaySound = row.operator(panel_tool.TUROK2_OT_play_sound_id.bl_idname, icon='PLAY', text="")
                    opPlaySound.soundID = item.get_args(i)
                    opPlaySound.soundIDIndex = -1
                    row.operator(panel_tool.TUROK2_OT_stop_all_sounds.bl_idname, icon='PAUSE', text="")
                    soundList = kdefs.defs["sounds"].get(item.get_args(i))
                    if soundList is not None and len(soundList) > 1:
                        if len(soundList) == 2:
                            secondText = os.path.splitext(os.path.split(soundList[1]["path"])[1])[0]
                        else:
                            secondText = "%i other sounds" % (len(soundList) - 1)
                        #end if
                        row = col.row(align=True)
                        row.alignment = "RIGHT"
                        row.label(text=secondText)
                        row.label(text="", icon='BLANK1')
                        row.label(text="", icon='BLANK1')
                    #end if
                elif argDesc == "Particle ID":
                    row.operator(TUROK2_OT_anim_action_arg_particle_list.bl_idname, icon='COLLAPSEMENU', text="").argIndex = i
                    col.alignment = "RIGHT"
                    col.label(text=argText)
                elif argDesc == "Voice ID":
                    row.operator(TUROK2_OT_anim_action_arg_voice_list.bl_idname, icon='COLLAPSEMENU', text="").argIndex = i
                    row = col.row(align=True)
                    row.alignment = "RIGHT"
                    row.label(text=argText)
                    row.operator(panel_tool.TUROK2_OT_play_voice.bl_idname, icon='PLAY', text="").voiceID = item.get_args(i)
                    row.operator(panel_tool.TUROK2_OT_stop_all_sounds.bl_idname, icon='PAUSE', text="")
                elif argDesc == "Actor ID":
                    row.operator(TUROK2_OT_anim_action_arg_actor_list.bl_idname, icon='COLLAPSEMENU', text="").argIndex = i
                    row = col.row(align=True)
                    row.alignment = "RIGHT"
                    row.label(text=argText)
                elif argDesc == "Hot Point Bit":
                    row.operator(TUROK2_OT_anim_action_arg_hotpoint_list.bl_idname, icon='COLLAPSEMENU', text="").argIndex = i
                    row = col.row(align=True)
                    row.alignment = "RIGHT"
                    hpName = kdefs.defs["hotpoints"].get(item.get_args(i), '')
                    row.label(text=hpName)
                elif argDesc == "Impact Type":
                    row.operator(TUROK2_OT_anim_action_arg_impact_list.bl_idname, icon='COLLAPSEMENU', text="").argIndex = i
                    row = col.row(align=True)
                    row.alignment = "RIGHT"
                    row.label(text=argText)
                elif argDesc == "Damage":
                    row = col.row(align=True)
                    row.alignment = "RIGHT"
                    row.label(text="Actual Damage: %f" % (item.get_args(i) / 256.0))
                elif item.event == 22 and i == 0: #User/Script Event arg1
                    if len(kdefs.defs["user_events"]["events"]) > 0:
                        row.operator(TUROK2_OT_anim_action_arg_userevent1_list.bl_idname, icon='COLLAPSEMENU', text="").argIndex = i
                    #end if
                elif item.event == 22 and i > 0: #User/Script Event arg2 or more
                    listInfo = kdefs.defs["user_events"]["lists"].get(argDesc)
                    if listInfo is not None:
                        opUserEventList = row.operator(TUROK2_OT_anim_action_arg_userevent_list.bl_idname, icon='COLLAPSEMENU', text="")
                        opUserEventList.argIndex = i
                        opUserEventList.listName = argDesc
                        row = col.row(align=True)
                        row.alignment = "RIGHT"
                        row.label(text=argText)
                    #end if
                else:
                    if argText != "":
                        row = col.row(align=True)
                        row.alignment = "RIGHT"
                        row.label(text=argText)
                    #end if
                #end if
            #end for
            
            #KeyFrameEvent Point options
            box.separator()
            titleRow = box.box().row()
            titleRow.alignment = "CENTER"
            titleRow.label(text="Point", icon="SHADING_SOLID")

            row = box.row(align=True)
            row.alignment = "CENTER"
            row.prop(obj.t2, "kfaPointIsRoot")
            turok2PointObj = utils.get_object(utils.TUROK2_ACTIONPOINT_NAME)
            pointIcon = "KEYTYPE_EXTREME_VEC" if turok2PointObj is None else "X"
            pointText = "Show Point" if turok2PointObj is None else "Remove Point"
            row = box.row(align=True)
            row.operator(TUROK2_OT_key_frame_toggle_point.bl_idname, icon=pointIcon, text=pointText)
            row.operator(TUROK2_OT_key_frame_point_update.bl_idname, icon="EXPORT", text="Update Point")
            row.operator(TUROK2_OT_key_frame_point_to_args.bl_idname, icon="IMPORT", text="Set Offset")
            row = box.row(align=True)
            row.prop(item, "offset")
        #end if
    #end if
#end def
# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(Turok2ActionKeyFramesPropGroup)
    bpy.utils.register_class(TUROK2_UL_key_frames_props)
    bpy.utils.register_class(Turok2AnimationActionSettings)
    bpy.utils.register_class(TUROK2_OT_loop_frame_last)
    bpy.utils.register_class(TUROK2_OT_key_frame_action)
    bpy.utils.register_class(TUROK2_OT_key_frame_copy)
    bpy.utils.register_class(TUROK2_OT_key_frame_paste)
    bpy.utils.register_class(TUROK2_OT_key_frame_set_anim_frame)
    bpy.utils.register_class(TUROK2_OT_key_frame_toggle_point)
    bpy.utils.register_class(TUROK2_OT_key_frame_point_update)
    bpy.utils.register_class(TUROK2_OT_key_frame_point_to_args)
    bpy.utils.register_class(TUROK2_MT_key_frame_special_menu)
    bpy.utils.register_class(TUROK2_OT_action_bone_list)
    bpy.utils.register_class(TUROK2_OT_anim_event_list)
    bpy.utils.register_class(TUROK2_OT_anim_action_arg_sound_list)
    bpy.utils.register_class(TUROK2_OT_anim_action_arg_voice_list)
    bpy.utils.register_class(TUROK2_OT_anim_action_arg_particle_list)
    bpy.utils.register_class(TUROK2_OT_anim_action_arg_actor_list)
    bpy.utils.register_class(TUROK2_OT_anim_action_arg_hotpoint_list)
    bpy.utils.register_class(TUROK2_OT_anim_action_arg_impact_list)
    bpy.utils.register_class(TUROK2_OT_anim_list)
    bpy.utils.register_class(TUROK2_OT_anim_action_arg_userevent_list)
    bpy.utils.register_class(TUROK2_OT_anim_action_arg_userevent1_list)
    bpy.utils.register_class(TUROK2_OT_key_frame_set_bone_none)
    bpy.utils.register_class(TUROK2_PT_action)
    bpy.types.Action.t2 = bpy.props.PointerProperty(type=Turok2AnimationActionSettings)
#end def

# -----------------------------------------------------------------------------
#     
def unregister():
    bpy.utils.unregister_class(Turok2ActionKeyFramesPropGroup)
    bpy.utils.unregister_class(TUROK2_UL_key_frames_props)
    bpy.utils.unregister_class(Turok2AnimationActionSettings)
    bpy.utils.unregister_class(TUROK2_OT_loop_frame_last)
    bpy.utils.unregister_class(TUROK2_OT_key_frame_action)
    bpy.utils.unregister_class(TUROK2_OT_key_frame_copy)
    bpy.utils.unregister_class(TUROK2_OT_key_frame_paste)
    bpy.utils.unregister_class(TUROK2_OT_key_frame_set_anim_frame)
    bpy.utils.unregister_class(TUROK2_OT_key_frame_toggle_point)
    bpy.utils.unregister_class(TUROK2_OT_key_frame_point_update)
    bpy.utils.unregister_class(TUROK2_OT_key_frame_point_to_args)
    bpy.utils.unregister_class(TUROK2_MT_key_frame_special_menu)
    bpy.utils.unregister_class(TUROK2_OT_action_bone_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_event_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_action_arg_sound_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_action_arg_voice_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_action_arg_particle_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_action_arg_actor_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_action_arg_hotpoint_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_action_arg_impact_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_action_arg_userevent_list)
    bpy.utils.unregister_class(TUROK2_OT_anim_action_arg_userevent1_list)
    bpy.utils.unregister_class(TUROK2_OT_key_frame_set_bone_none)
    bpy.utils.unregister_class(TUROK2_PT_action)
    del bpy.types.Action.t2
#end def
