import bpy
import os
from datetime import (
    datetime,
    timedelta,
)

from . import (
    utils,
    file_utils,
    paths,
    kdefs,
)

from bpy.types import Operator, AddonPreferences

PREF_SAVE_IGNORE_KEYS = ["isInitialized"]

# -----------------------------------------------------------------------------
#
class Turok2AddonPreferences(AddonPreferences):
    bl_idname = __package__
    
    def on_get_gamedir(self):
        return self.get("gameDirectory", "")
    #end if
    def on_set_gamedir(self, value):
        self.try_set_var(value, "gameDirectory")
    #end if
    
    def on_get_devdir(self):
        return self.get("devDirectory", "")
    #end if
    def on_set_devdir(self, value):
        self.try_set_var(value, "devDirectory")
    #end if
    
    def on_get_fps60playback(self):
        return self.get("fps60playback", True)
    #end if
    def on_set_fps60playback(self, value):
        self.try_set_var(value, "fps60playback")
    #end if

    def on_get_animAutoName(self):
        return self.get("animAutoName", True)
    #end if
    def on_set_animAutoName(self, value):
        self.try_set_var(value, "animAutoName")
    #end if

    def on_get_hideKFAArgs(self):
        return self.get("hideKFAArgs", True)
    #end if
    def on_set_hideKFAArgs(self, value):
        self.try_set_var(value, "hideKFAArgs")
    #end if

    def try_set_var(prefs, value, key):
        if key not in prefs or prefs[key] != value:
            prefs[key] = value
            file_utils.save_class_to_json(paths.custom_prefs(), utils.prefs(), PREF_SAVE_IGNORE_KEYS)
            return True
        #end if
        
        return False
    #end if

    isInitialized: bpy.props.BoolProperty(name="Is Initialized", default=False)
    gameDirectory: bpy.props.StringProperty(name="Unzipped Game Folder", default="", subtype='DIR_PATH', get=on_get_gamedir, set=on_set_gamedir)
    devDirectory: bpy.props.StringProperty(name="Turok 2 Folder", default="", subtype='DIR_PATH', get=on_get_devdir, set=on_set_devdir)
    fps60playback: bpy.props.BoolProperty(name="60 FPS Playback", description="Force Skinnedmesh Animations to play at 60 FPS", default=True, get=on_get_fps60playback, set=on_set_fps60playback)
    animAutoName: bpy.props.BoolProperty(name="Auto name Animations", description="When you change the animation ID it will change the name of the animation automatically", default=True, get=on_get_animAutoName, set=on_set_animAutoName)
    hideKFAArgs: bpy.props.BoolProperty(name="Hide Unused Key Frame Action Args", description="Hides the Args that are unused for the Key Frame Action Events", default=True, get=on_get_hideKFAArgs, set=on_set_hideKFAArgs)

    def draw(self, context):
        layout = self.layout
        row = layout.row(align=True)
        #row.alignment = 'RIGHT'
        row.operator(TUROK2_OT_pref_open_addon_folder.bl_idname, icon='FILE_FOLDER', text="Open Add-on Folder")
        row.label(text="")
        row.operator(TUROK2_OT_pref_open_website_link.bl_idname, icon='URL', text="Go To Website")
        row.operator(TUROK2_OT_pref_copy_website_link.bl_idname, icon='COPYDOWN', text="")
        row = layout.row()
        # layout.box()
        
        box = layout.box()
        row = box.row()
        row.alignment = 'CENTER'
        row.label(text="Paths")
        if self.gameDirectory == "" or self.devDirectory == "":
            row = box.row()
            row.alert = True
            row.label(text="You must set the following paths before using this add-on", icon="ERROR")
        #end if
        row = box.row(align=True)
        if self.gameDirectory == "":
            row.alert = True
        #end if
        row.prop(self, "gameDirectory")
        row = box.row(align=True)
        if self.devDirectory == "":
            row.alert = True
        #end if
        row.prop(self, "devDirectory")
        
        #Load Defs
        box = layout.box()
        row = box.row()
        row.alignment = 'CENTER'
        row.label(text="Load Defs")
        row = box.row(align=True)
        baseDefsLoaded = len(kdefs.defs["anims"]) > 0 and len(kdefs.defs["anim_events"]) > 0 and len(kdefs.defs["impacts"]) > 0 and len(kdefs.defs["hotpoints"]) > 0 and len(kdefs.defs["builtin_kmats"]) > 0 and len(kdefs.defs["texsets"]) > 0 and len(kdefs.defs["voices"]) > 0
        allDefsLoaded = baseDefsLoaded and len(kdefs.defs["actors"]) > 0 and len(kdefs.defs["particles"]) > 0 and len(kdefs.defs["sounds"]) > 0
        row = row.row(align=True)
        row.alert = not allDefsLoaded
        opIcon = 'NONE' if row.alert else 'CHECKMARK'
        row.operator(TUROK2_OT_pref_load_defs_all.bl_idname, text="All", icon=opIcon)
        row = row.row(align=True)
        row.alert = not baseDefsLoaded
        opIcon = 'NONE' if row.alert else 'CHECKMARK'
        row.operator(TUROK2_OT_pref_load_defs_base.bl_idname, text="Base", icon=opIcon)
        row = row.row(align=True)
        row.alert = len(kdefs.defs["actors"]) == 0
        opIcon = 'NONE' if row.alert else 'CHECKMARK'
        row.operator(TUROK2_OT_pref_load_defs_actors.bl_idname, text="Actors", icon=opIcon)
        row = row.row(align=True)
        row.alert = len(kdefs.defs["particles"]) == 0
        opIcon = 'NONE' if row.alert else 'CHECKMARK'
        row.operator(TUROK2_OT_pref_load_defs_particles.bl_idname, text="Particles", icon=opIcon)
        row = row.row(align=True)
        row.alert = len(kdefs.defs["sounds"]) == 0
        opIcon = 'NONE' if row.alert else 'CHECKMARK'
        row.operator(TUROK2_OT_pref_load_defs_sounds.bl_idname, text="Sounds", icon=opIcon)
        
        #Misc Settings
        box = layout.box()
        row = box.row()
        row.alignment = 'CENTER'
        row.label(text="Settings")
        row = box.row(align=True)
        # row.alignment = "LEFT"
        row.prop(self, "fps60playback")
        row.prop(self, "animAutoName")
        row = box.row(align=True)
        row.prop(self, "hideKFAArgs")
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_pref_open_addon_folder(bpy.types.Operator):
    bl_idname = "turok2_pref.open_addon_folder"
    bl_label = "Open Turok 2 addon folder"
    bl_description = "Tries to open the Turok 2 addon folder"
    
    def execute(self, context):
        file_utils.open_folder(paths.addon_dir())
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_pref_open_website_link(bpy.types.Operator):
    bl_idname = "turok2_pref.open_wesbite_link"
    bl_label = "Open Turok 2 Tools Website"
    bl_description = "Open link in web browser to Turok 2 EX Tools Website"
    bl_options = {'INTERNAL'}
    
    def execute(self, context):
        utils.open_update_link(utils.TUROK2_WEBSITE)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_pref_copy_website_link(bpy.types.Operator):
    bl_idname = "turok2_pref.copy_website_link"
    bl_label = "Copy website link"
    bl_description = "Copies the website link to the clipboard"
    bl_options = {'INTERNAL'}
    
    def execute(self, context):
        bpy.context.window_manager.clipboard = utils.TUROK2_WEBSITE
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_pref_load_defs_all(bpy.types.Operator):
    bl_idname = "turok2_pref.load_defs_all"
    bl_label = "Load All Def Files"
    bl_description = "Loads all def,actor,sounds,particles,etc files from the game"
    
    def execute(self, context):
        kdefs.load_all()
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_pref_load_defs_base(bpy.types.Operator):
    bl_idname = "turok2_pref.load_defs_base"
    bl_label = "Load Base Def Files"
    bl_description = "Loads base def files. Both builtin and game def files."
    
    def execute(self, context):
        kdefs.load_defs_base()
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_pref_load_defs_actors(bpy.types.Operator):
    bl_idname = "turok2_pref.load_defs_actors"
    bl_label = "Load Actor Def Files"
    bl_description = "Loads actor def files"
    
    def execute(self, context):
        kdefs.load_defs_actors()
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_pref_load_defs_particles(bpy.types.Operator):
    bl_idname = "turok2_pref.load_defs_particles"
    bl_label = "Load Particle Def Files"
    bl_description = "Loads particle def files"
    
    def execute(self, context):
        kdefs.load_defs_particles()
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_pref_load_defs_sounds(bpy.types.Operator):
    bl_idname = "turok2_pref.load_defs_sounds"
    bl_label = "Load Sound Def Files"
    bl_description = "Loads sound def files"
    
    def execute(self, context):
        kdefs.load_defs_sounds()
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(Turok2AddonPreferences)
    bpy.utils.register_class(TUROK2_OT_pref_open_addon_folder)
    bpy.utils.register_class(TUROK2_OT_pref_open_website_link)
    bpy.utils.register_class(TUROK2_OT_pref_copy_website_link)
    bpy.utils.register_class(TUROK2_OT_pref_load_defs_all)
    bpy.utils.register_class(TUROK2_OT_pref_load_defs_base)
    bpy.utils.register_class(TUROK2_OT_pref_load_defs_actors)
    bpy.utils.register_class(TUROK2_OT_pref_load_defs_particles)
    bpy.utils.register_class(TUROK2_OT_pref_load_defs_sounds)
    
    #Load custom preferences if addon was enabled
    if not utils.prefs().isInitialized:
        utils.prefs().isInitialized = True
        file_utils.load_json_to_class(paths.custom_prefs(), utils.prefs(), PREF_SAVE_IGNORE_KEYS)
    #end if
#end def

# -----------------------------------------------------------------------------
#     
def unregister():
    bpy.utils.unregister_class(Turok2AddonPreferences)
    bpy.utils.unregister_class(TUROK2_OT_pref_open_addon_folder)
    bpy.utils.unregister_class(TUROK2_OT_pref_open_website_link)
    bpy.utils.unregister_class(TUROK2_OT_pref_copy_website_link)
    bpy.utils.unregister_class(TUROK2_OT_pref_load_defs_all)
    bpy.utils.unregister_class(TUROK2_OT_pref_load_defs_base)
    bpy.utils.unregister_class(TUROK2_OT_pref_load_defs_actors)
    bpy.utils.unregister_class(TUROK2_OT_pref_load_defs_particles)
    bpy.utils.unregister_class(TUROK2_OT_pref_load_defs_sounds)
#end def
