import os
from . import (
    utils,
    file_utils,
)

# -----------------------------------------------------------------------------
# Returns the path to the this addons directory
def addon_dir():
    return os.path.dirname(os.path.realpath(__file__))
#end def

# -----------------------------------------------------------------------------
# Returns the path to the this addons textures directory
def addon_texture_dir():
    return os.path.join(addon_dir(), "textures")
#end def

# -----------------------------------------------------------------------------
# returns the file path to the animations def file
def def_anims():
    return os.path.join(addon_dir(), "defs", "animations.txt")
#end def

# -----------------------------------------------------------------------------
# returns the file path to the animevents def file
def def_animevents():
    return os.path.join(addon_dir(), "defs", "animevents.txt")
#end def

# -----------------------------------------------------------------------------
# returns the file path to the impact types def file
def def_impacts():
    return os.path.join(addon_dir(), "defs", "impacts.txt")
#end def

# -----------------------------------------------------------------------------
# returns the file path to the hot points def file
def def_hotpoints():
    return os.path.join(addon_dir(), "defs", "hotpoints.txt")
#end def

# -----------------------------------------------------------------------------
# returns the file path to the hot points def file
def def_builtin_kmats():
    return os.path.join(addon_dir(), "defs", "builtin.kmat")
#end def

# -----------------------------------------------------------------------------
# returns the file path to the custom preferences save file
def custom_prefs():
    return os.path.join(addon_dir(), "saves", "prefs.txt")
#end def

# -----------------------------------------------------------------------------
# returns the file path to the custom def cache save file
def custom_defcache():
    return os.path.join(addon_dir(), "saves", "defcache.pickle")
#end def

# -----------------------------------------------------------------------------
# Returns a set of absolute file paths from the dev/game directory. if dev path already exists the game path is not added.
# dirPath: sub directory in dev and game directory to search for files
# ext: str or tuple of str eg. ".txt" or (".txt", ".json")
def find_t2_files_in_dir(dirPath, ext=''):
    devDir = utils.prefs().devDirectory
    gameDir = utils.prefs().gameDirectory
    devFiles = file_utils.get_files_in_dir_relpath(os.path.join(devDir, dirPath), devDir, ext)
    gameFiles = file_utils.get_files_in_dir_relpath(os.path.join(gameDir, dirPath), gameDir, ext)
    finalFiles = {os.path.join(devDir, devFile) for devFile in devFiles}
    for gameFile in gameFiles:
        if gameFile not in devFiles:
            finalFiles.add(os.path.join(gameDir, gameFile))
        #end if
    #end for
    
    return finalFiles
#end def

# # -----------------------------------------------------------------------------
# # Returns 2 sets of relative file paths from the dev and game directory. if a dev path already exists the game path is not added to it.
# # dirPath: sub directory in dev and game directory to search for files
# # ext: str or tuple of str eg. ".txt" or (".txt", ".json")
# def find_t2_files_in_dir_devgame(dirPath, ext=''):
    # devDir = utils.prefs().devDirectory
    # gameDir = utils.prefs().gameDirectory
    # devFiles = file_utils.get_files_in_dir_relpath(os.path.join(devDir, dirPath), devDir, ext)
    # gameFiles = file_utils.get_files_in_dir_relpath(os.path.join(gameDir, dirPath), gameDir, ext)
    # uniqueGameFiles = {gameFile for gameFile in gameFiles if gameFile not in devFiles}
    # # uniqueGameFiles = set()
    # # for gameFile in gameFiles:
        # # if gameFile not in devFiles:
            # # finalFiles.add(os.path.join(gameDir, gameFile))
        # # #end if
    # # #end for
    
    # return devFiles, uniqueGameFiles
# #end def

# -----------------------------------------------------------------------------
# 
def find_t2_file_path(relPath):
    absPaths = [os.path.join(utils.prefs().devDirectory, relPath), os.path.join(utils.prefs().gameDirectory, relPath)]
    for absPath in absPaths:
        absPath = os.path.normpath(absPath)
        if os.path.isfile(absPath):
            return absPath
        #end if
    #end for

    return None
#end def

# -----------------------------------------------------------------------------
# Returns the local game texture path (sampler path) from the absolute texture path
def texture_filepath_to_local_path(filepath, startWithTextures=True):
    newFilePath = os.path.normpath(filepath).strip('\\')
    texFileDir, texfileName = os.path.split(newFilePath)
    texfileNameNoExt = os.path.splitext(texfileName)[0]

    if texfileNameNoExt in utils.TUROK2_BUILTIN_TEXTURES:
        return texfileNameNoExt
    else:
        parentTexturesDir, texturesSubDir = get_path_from_dir(texFileDir, "textures")
        #if found a textures sub directory in file path
        if parentTexturesDir != texFileDir and texturesSubDir != "":
            if startWithTextures:
                return "textures/%s/%s" % (texturesSubDir, texfileName)
            else:
                return "%s/%s" % (texturesSubDir, texfileName)
            #end if
        else:
            if startWithTextures:
                return "textures/%s" % (texfileName)
            else:
                return texfileName
            #end if
        #end if
    #end if
#end def

# -----------------------------------------------------------------------------
# Returns a path by finding a folder named folderName in fullDir and returning it's parent path.
# If folderName is not found returns the fullDir.
# Second Return is the path after the folderName or fullDir if not found.
def get_path_from_dir(fullDir, folderName):
    path = fullDir
    path2 = ""
    while True:
        path, name = os.path.split(path)
        if name == folderName:
            return path, path2
        elif path == "" or name == "":
            break
        #end if
        if path2 == "":
            path2 = name
        else:
            path2 = os.path.join(name, path2)
        #end if
    #end while
    
    return fullDir, fullDir
#end def

# -----------------------------------------------------------------------------
# Returns the dir path in the form of kmat string where each folder is separated by a /
def GetLocalKexPathFromDir(dir):
    result = ""
    path = dir
    while True:
        path, name = os.path.split(path)
        if name == "":
            break
        #end if
        if result == "":
            result = name
        else:
            result = "%s/%s" % (name, result)
        #end if
    #end while
    
    return result
#end def
